# -*- coding: utf-8 -*-

import importlib


def load_module_from_app(app: str, dotted_path: str):
    """Load module from application."""
    module_name = "{0}.{1}".format(app, dotted_path)
    module_spec = importlib.util.find_spec(module_name)  # type: ignore

    if module_spec:
        return importlib.import_module(module_name)



# -*- coding: utf-8 -*-

from functools import wraps

from django.forms.utils import pretty_name
from django.utils.html import format_html

from jnt_django_toolbox.admin.helpers import urls
from jnt_django_toolbox.helpers.objects import getattr_nested


def admin_changelist_link(
    attr, short_description=None, empty_description="-", query_string=None,
):
    """Render a link to the list of a related model in the admin changelist.

    ``attr (str)``
        Name of the related field.
    ``short_description (str)``
        Field display name.
        Default value: None.
    ``empty_description (str)``
        Value to display if the related field is None.
        Default value: -.
    ``query_string (function)``
        Optional callback for adding a query string to the link.
        Receives the object and should return a query string.
        Default value: None.

    The wrapped method receives the related object and
    should return the link text.

    Usage::

        from jnt_admin_tools.decorators import admin_changelist_link
        from django.contrib import admin
        from test_app.models import Bar

        @admin.register(Bar)
        class BarAdmin(admin.ModelAdmin):
            fields = ('name', 'foos', 'custom_field')
            readonly_fields = ('foos', 'custom_field')

            @admin_changelist_link(
                'foos',
                query_string=lambda bar: 'bar_id={0}'.format(bar.pk)
            )
            def foos(self, foos):
                return ', '.join(str(foo) for foo in foos.all())

    .. image:: images/decorators/decorator_admin_changelist_link.png
    """

    def wrap(func):  # noqa: WPS430
        @wraps(func)
        def field_func(self, obj):  # noqa: WPS430
            related_obj = getattr_nested(obj, attr)
            if related_obj is None:
                return empty_description
            url = urls.admin_changelist_url(related_obj.model)
            if query_string:
                url = "{0}?{1}".format(url, query_string(obj))
            return format_html(
                '<a href="{0}">{1}</a>', url, func(self, related_obj),
            )

        field_func.short_description = (
            short_description if short_description else pretty_name(attr)
        )
        field_func.allow_tags = True
        return field_func

    return wrap



# -*- coding: utf-8 -*-

from datetime import date, datetime, timedelta
from math import ceil
from typing import Union

from jnt_django_toolbox.consts.time import SECONDS_PER_HOUR, SECONDS_PER_MINUTE

Number = Union[int, float]


def date2datetime(value: date) -> datetime:  # noqa: WPS110
    """Converts date to datetime."""
    return datetime.combine(value, datetime.min.time())


def begin_of_week(value: date) -> date:  # noqa: WPS110
    """Get begin of week."""
    return value - timedelta(days=value.weekday() % 7)


def humanize_time(total_seconds: Number) -> str:
    """Convert seconds to humanize_time."""
    if not isinstance(total_seconds, (int, float)):
        raise ValueError("Seconds should be a number")

    hours, seconds = divmod(ceil(total_seconds), SECONDS_PER_HOUR)
    minutes, seconds = divmod(seconds, SECONDS_PER_MINUTE)

    time_parts = [minutes, seconds]
    if hours:
        time_parts.insert(0, hours)

    return ":".join(("{0:02d}".format(part) for part in time_parts))



# -*- coding: utf-8 -*-

from contextlib import suppress

from django.core.exceptions import ImproperlyConfigured
from six import string_types


def cmp(a, b):
    return (a > b) - (a < b)


class Bit:
    """
    Represents a single Bit.
    """

    def __init__(self, number, is_set=True):
        self.number = number
        self.is_set = bool(is_set)
        self.mask = 2 ** int(number)
        self.children = []
        if not self.is_set:
            self.mask = ~self.mask

    def evaluate(self, evaluator, qn, connection):
        return self.mask, []

    def prepare(self, evaluator, query, allow_joins):
        return evaluator.prepare_node(self, query, allow_joins)

    def __repr__(self):
        return "<%s: number=%d, is_set=%s>" % (
            self.__class__.__name__,
            self.number,
            self.is_set,
        )

    def __int__(self):
        return self.mask

    def __bool__(self):
        return self.is_set

    def __eq__(self, value):
        if isinstance(value, Bit):
            return value.number == self.number and value.is_set == self.is_set
        elif isinstance(value, bool):
            return value == self.is_set
        elif isinstance(value, int):
            return value == self.mask
        return value == self.is_set

    def __ne__(self, value):
        return not self == value  # noqa: WPS508

    def __coerce__(self, value):
        return (self.is_set, bool(value))

    def __invert__(self):
        return self.__class__(self.number, not self.is_set)

    def __and__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return value & self.mask

    def __rand__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return self.mask & value

    def __or__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return value | self.mask

    def __ror__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return self.mask | value

    def __lshift__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return value << self.mask

    def __rlshift__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return self.mask << value

    def __rshift__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return value >> self.mask

    def __rrshift__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return self.mask >> value

    def __xor__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return value ^ self.mask

    def __rxor__(self, value):
        if isinstance(value, Bit):
            value = value.mask
        return self.mask ^ value


class BitHandler:
    """
    Represents an array of bits, each as a ``Bit`` object.
    """

    def __init__(self, value, keys, labels=None):
        # TODO: change to bitarray?
        if value:
            self._value = int(value)
        else:
            self._value = 0
        self._keys = keys
        self._labels = labels is not None and labels or keys

    def __str__(self):
        return str(self._value)

    @property
    def mask(self):
        return self._value

    def evaluate(self, evaluator, qn, connection):
        return self.mask, []

    def get_bit(self, bit_number):
        mask = 2 ** int(bit_number)
        return Bit(bit_number, self._value & mask != 0)

    def set_bit(self, bit_number, true_or_false):
        mask = 2 ** int(bit_number)
        if true_or_false:
            self._value |= mask
        else:
            self._value &= ~mask
        return Bit(bit_number, self._value & mask != 0)

    def keys(self):
        return self._keys

    def items(self):
        return [(key, getattr(self, key).is_set) for key in self._keys]

    def get_label(self, flag):
        if isinstance(flag, string_types):
            flag = self._keys.index(flag)
        if isinstance(flag, Bit):
            flag = flag.number
        return self._labels[flag]

    def __eq__(self, other):
        if not isinstance(other, BitHandler):
            return False
        return self._value == other._value

    def __lt__(self, other):
        return int(self._value) < other

    def __le__(self, other):
        return int(self._value) <= other

    def __gt__(self, other):
        return int(self._value) > other

    def __ge__(self, other):
        return int(self._value) >= other

    def __cmp__(self, other):
        return cmp(self._value, other)

    def __repr__(self):
        return "<{0}: {1}>".format(
            self.__class__.__name__,
            ", ".join(
                "{0}={1}".format(k, self.get_bit(n).is_set)
                for n, k in enumerate(self._keys)
            ),
        )

    def __int__(self):
        return self._value

    def __bool__(self):
        return bool(self._value)

    def __and__(self, value):
        return BitHandler(self._value & int(value), self._keys)

    def __or__(self, value):
        return BitHandler(self._value | int(value), self._keys)

    def __add__(self, value):
        return BitHandler(self._value + int(value), self._keys)

    def __sub__(self, value):
        return BitHandler(self._value - int(value), self._keys)

    def __lshift__(self, value):
        return BitHandler(self._value << int(value), self._keys)

    def __rshift__(self, value):
        return BitHandler(self._value >> int(value), self._keys)

    def __xor__(self, value):
        return BitHandler(self._value ^ int(value), self._keys)

    def __contains__(self, key):
        bit_number = self._keys.index(key)
        return bool(self.get_bit(bit_number))

    def __getattr__(self, key):
        if key.startswith("_"):
            return object.__getattribute__(self, key)
        if key not in self._keys:
            raise AttributeError("{0} is not a valid flag".format(key))
        return self.get_bit(self._keys.index(key))

    def __setattr__(self, key, value):
        if key.startswith("_"):
            return object.__setattr__(self, key, value)
        if key not in self._keys:
            raise AttributeError("{0} is not a valid flag".format(key))
        self.set_bit(self._keys.index(key), value)

    def __iter__(self):
        yield from self.items()


def register_sqlite3_adapters():
    with suppress(ImproperlyConfigured):
        from django.db.backends.sqlite3.base import Database

        Database.register_adapter(Bit, int)
        Database.register_adapter(BitHandler, int)


def register_postgres_adapters():
    with suppress(ImproperlyConfigured):
        from django.db.backends.postgresql.base import Database

        Database.extensions.register_adapter(
            Bit, lambda x: Database.extensions.AsIs(int(x)),
        )
        Database.extensions.register_adapter(
            BitHandler, lambda x: Database.extensions.AsIs(int(x)),
        )


register_sqlite3_adapters()
register_postgres_adapters()



# -*- coding: utf-8 -*-

import collections

not_provided = object()


def recursive_dict():
    """Default dict of default dicts with unlimited nesting."""
    return collections.defaultdict(recursive_dict)


def deep_get(dic, path, default=not_provided):
    """Get a nested dict value by dot-separated keys."""
    keys = path.split(".")

    try:  # noqa:WPS229
        for key in keys[:-1]:
            dic = dic[key]

        return dic[keys[-1]]
    except KeyError:
        if default == not_provided:
            raise

        return default


def deep_set(dic, path, value_to_set):
    """Set a nested dict value by dot-separated keys."""
    keys = path.split(".")

    for key in keys[:-1]:
        dic = dic[key]

    dic[keys[-1]] = value_to_set



# -*- coding: utf-8 -*-

from functools import reduce
from typing import Dict


def getattr_nested(instance, attr, *args):
    """Get object attribute."""
    try:
        return reduce(getattr, [instance] + attr.split("__"))
    except AttributeError:
        if not len(args):
            raise

        return args[0]


class ObjectView:
    """Object view."""

    def __init__(self, dictionary: Dict[str, object]):
        """Initialize self."""
        self.__dict__ = dictionary


def dict2obj(dictionary: Dict[str, object]) -> ObjectView:
    """Create ObjectView from dictionary."""
    return ObjectView(dictionary)



# -*- coding: utf-8 -*-

from django.forms import CheckboxSelectMultiple
from django.utils.encoding import force_text

from jnt_django_toolbox.models.fields.bit.types import BitHandler


class BitFieldWidget(CheckboxSelectMultiple):
    def render(self, name, value, attrs=None, choices=(), renderer=None):
        if isinstance(value, BitHandler):
            value = [k for k, v in value if v]
        elif isinstance(value, int):
            real_value = []
            div = 2
            for k, _ in self.choices:
                if value % div != 0:
                    real_value.append(k)
                    value -= value % div
                div *= 2
            value = real_value
        return super().render(name, value, attrs=attrs)

    def _has_changed(self, initial, data):
        if initial is None:
            initial = []
        if data is None:
            data = []
        if initial != data:
            return True
        initial_set = {force_text(value) for value in initial}
        data_set = {force_text(value) for value in data}
        return data_set != initial_set



# -*- coding: utf-8 -*-

from functools import wraps


def admin_field(short_description=None, allow_tags=True):
    """Render custom field or add link in the admin detail page.

    ``short_description (str)``
        Description of the field.
    ``allow_tags (bool)``
        Allow tags.

    The wrapped method receives the custom field.

    Usage::

        from jnt_admin_tools.decorators import admin_field
        from django.contrib import admin
        from django.utils.safestring import mark_safe
        from test_app.models import Bar

        @admin.register(Bar)
        class BarAdmin(admin.ModelAdmin):
            fields = ('name', 'foos', 'custom_field')
            readonly_fields = ('foos', 'custom_field')

            @admin_field('Custom field')
            def custom_field(self, obj):
                return mark_safe('<h1>custom field</h1>')

    .. image:: images/decorators/decorator_admin_field.png
    """

    def wrap(func):  # noqa: WPS430
        @wraps(func)
        def field_func(self, instance):  # noqa: WPS430
            return func(self, instance)

        field_func.short_description = short_description
        field_func.allow_tags = allow_tags
        return field_func

    return wrap



# -*- coding: utf-8 -*-

from contextlib import suppress

from django.urls import NoReverseMatch, reverse


def admin_change_url(instance):
    return reverse(
        "admin:{0}_{1}_change".format(
            instance._meta.app_label, instance._meta.model.__name__.lower(),
        ),
        args=(instance.pk,),
    )


def admin_changelist_url(model):
    return reverse(
        "admin:{0}_{1}_changelist".format(
            model._meta.app_label, model.__name__.lower(),
        ),
    )


def admin_autocomplete_url(model):
    with suppress(NoReverseMatch):
        return reverse(
            "admin:{0}_{1}_autocomplete".format(
                model._meta.app_label, model.__name__.lower(),
            ),
        )
